/*******************************************************************************
**                        backupprofile.cpp
**                             part of
**                konserve -- A small backup application
**			 -------------------
**  begin    : Mon Apr  8 23:19:44 CEST 2002
**  copyright: (C) 2002 by Florian Simnacher
**  email    : simnacher@gmx.de
**  $Id: backupprofile.cpp,v 1.3 2003/09/17 10:12:28 florian Exp $
*******************************************************************************/

/*******************************************************************************
**  This program is free software; you can redistribute it and/or             **
**  modify it under the terms of the GNU General Public License               **
**  as published by the Free Software Foundation; either version 2            **
**  of the License, or (at your option) any later version.                    **
**                                                                            **
**  This program is distributed in the hope that it will be useful,           **
**  but WITHOUT ANY WARRANTY; without even the implied warranty of            **
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             **
**  GNU General Public License for more details.                              **
**                                                                            **
**  You shougld have received a copy of the GNU General Public License        **
**  along with this program; if not, write to the Free Software               **
**  Foundation, Inc., 59 Temple Place - Suite 330, Boston,                    **
**  MA  02111-1307, USA.                                                      **
*******************************************************************************/

// Standard includes
#include <iostream>

// Qt includes
#include <qfile.h>
#include <qfileinfo.h>
#include <qstringlist.h>
#include <qdir.h>

// KDE includes
#include <ktempfile.h>
#include <kmessagebox.h>
#include <klocale.h>
#include <kfileitem.h>
#include <kapplication.h>

// My includes
#include "backupprofile.h"
#include "helper.h"

////////////////////////////////////////////////////////////////////////////////
///////                  class Backup                                    ///////
////////////////////////////////////////////////////////////////////////////////

Backup::Backup( KURL surl,
                KURL aurl,
                bool readonly )
    : Dumpable( "Backup" /* , QString().setNum( (long) this, 16 ) */ ),
      mSourceUrl( surl ),
      mTmpSourceUrl(),
      mArchiveUrl( aurl ),
      mTmpArchiveUrl(),
      mbReadOnly( readonly ),
      mpArchive( 0L )
{
    // NOOP
}

Backup::~Backup()
{
    // NOOP
}

void Backup::doBackup()
{
    // std::cerr << "Backup::doBackup" << std::endl; //xxx
    Q_ASSERT( ! mSourceUrl.isMalformed() );
    Q_ASSERT( ! mArchiveUrl.isMalformed() );
    Q_ASSERT( ! mbReadOnly );

    KTempFile tmpFile;
    mTmpSourceUrl.setPath( tmpFile.name() );
    tmpFile.unlink();

    if ( mSourceUrl.isLocalFile() )
    {
        mTmpSourceUrl = mSourceUrl;
        slotStartListing( 0L );
    }
    else
    {
        KIO::Job* job = KIO::copy( mSourceUrl, mTmpSourceUrl );
        connect(
            dynamic_cast<const QObject*>(job), SIGNAL( result( KIO::Job * ) ),
            SLOT( slotStartListing( KIO::Job *) )
            );
    }
}

QString Backup::dumpInternal()
{
    QString tmp;

    tmp += QString( "mSourceUrl:  " ) + mSourceUrl.url()  + "\n";
    tmp += QString( "mArchiveUrl: " ) + mArchiveUrl.url() + "\n";
    if ( mbReadOnly )
        tmp += QString( "readonly" );
    else
        tmp += QString( "read/write" );

    return tmp;
}

void Backup::slotStartListing( KIO::Job* aJob)
{
    // std::cerr << "Backup::slotStartListing" << std::endl;
    Q_ASSERT( 0L == mpArchive );
    Q_ASSERT( mTmpSourceUrl.isValid() );
    Q_ASSERT( ! mTmpSourceUrl.isMalformed() );
    Q_ASSERT( mTmpSourceUrl.isLocalFile() );

    if ( 0L != aJob )
    {
        if ( aJob->error() )
        {
            aJob->showErrorDialog();
            return;
        }
    }
    KTempFile tmpFile;
    mTmpArchiveUrl.setPath( tmpFile.name() );
    tmpFile.unlink();

    mpArchive = new KTar( mTmpArchiveUrl.path(), "application/x-gzip" );
    mpArchive->open( IO_WriteOnly );
//     if ( QFileInfo( mTmpSourceUrl.path() ).isDir() )
//     {
//         mpArchive->writeDir(
//             mSourceUrl.path(), Helper::userName(), Helper::userGroup()
//             );
//     }
    KIO::Job* job = KIO::listRecursive( mTmpSourceUrl, false, true );
    connect(
        job, SIGNAL( entries( KIO::Job*, const KIO::UDSEntryList&)),
             SLOT( slotGetEntriesAndPackEm( KIO::Job*, const KIO::UDSEntryList&))
        );
    connect(
        job, SIGNAL( result( KIO::Job* ) ),
             SLOT( slotFinishedPacking( KIO::Job* ) )
        );

    Q_ASSERT( 0L != mpArchive );
    Q_ASSERT( mTmpArchiveUrl.isValid() );
    Q_ASSERT( ! mTmpArchiveUrl.isMalformed() );
    Q_ASSERT( mTmpArchiveUrl.isLocalFile() );
}

void Backup::slotGetEntriesAndPackEm( KIO::Job*, const KIO::UDSEntryList& list )
{
    // std::cerr << "Backup::slotGetEntriesAndPackEm" << std::endl;
    Q_ASSERT( mTmpSourceUrl.isValid() );
    Q_ASSERT( ! mTmpSourceUrl.isMalformed() );
    Q_ASSERT( mTmpSourceUrl.isLocalFile() );

    KIO::UDSEntryListConstIterator it=list.begin();
    for (; it != list.end(); ++it)
    {
        KFileItem f( *it, const_cast<const KURL&>( mTmpSourceUrl ) );

        if ( f.text() == "." || f.text() == ".." )
        {
            //std::cerr << "filter out . and .." << std::endl;
            continue;
        }
        if ( f.isDir() )
        {
            if ( !f.isReadable())
            {
                //std::cerr << f.url().path() << f.text() << "Not readable!" << std::endl;
                continue;
            }
            //std::cerr << f.url().url() << f.text() << std::endl;
            mpArchive->writeDir(
                // XXX hier haperts noch
                mSourceUrl.path() + f.text(), Helper::userName(),
                Helper::userGroup()
                );
        }
        else if ( f.isFile() || f.isLink() )
        {
            //std::cerr << f.url().path() << "/" << f.text() << std::endl;
            QFile *file = new QFile( f.url().path() + "/" + f.text() );
            Q_CHECK_PTR( file );
            Q_ASSERT( file->exists() );
            char  *c = new char[file->size()];

            if ( !file->open( IO_ReadOnly ) )
                continue;
            file->readBlock( c, file->size() );

            mpArchive->writeFile(
                // XXX hier haperts noch
                mSourceUrl.path() + f.text(), Helper::userName(),
                Helper::userGroup(), (unsigned int) file->size(), c
                );
            file->close();
            delete[] c;
            delete file;
        }
        else
        {
            //std::cerr << "shit: " << f.text() << " is nothing" << std::endl;
        }
//        kapp->processEvents();
    }
}

void Backup::slotFinishedPacking( KIO::Job* aJob )
{
    // std::cerr << "Backup::slotFinishedListing" << std::endl;
    Q_ASSERT( QFileInfo( mTmpArchiveUrl.path() ).exists() );
    if ( aJob->error() )
    {
        aJob->showErrorDialog();
        return;
    }
    mpArchive->close();
    // who has invented the protected base class Dtor???
    delete dynamic_cast<KTar*>(mpArchive);
    mpArchive = 0L;

    slotCopyArchive();
}

void Backup::slotCopyArchive()
{
    //std::cerr << "Backup::slotCopyArchive" << std::endl;
    Q_ASSERT( mArchiveUrl.isValid() );
    Q_ASSERT( ! mArchiveUrl.isMalformed() );
    Q_ASSERT( QFileInfo( mTmpArchiveUrl.path() ).exists() );
    Q_ASSERT( mTmpArchiveUrl.isValid() );
    Q_ASSERT( ! mTmpArchiveUrl.isMalformed() );
    Q_ASSERT( mTmpArchiveUrl.isLocalFile() );

    KURL destUrl;
    //std::cerr << "Archive: " << mArchiveUrl.path() << std::endl;
    if ( QFileInfo( mArchiveUrl.path() ).isDir() )
    {
        //std::cerr << "Backup::slotCopyArchive: isDir" << std::endl;
        destUrl = dotCheck( defaultArchiveName( mSourceUrl, mArchiveUrl ) );
    }
    else
    {
        destUrl = dotCheck( mArchiveUrl );
        //std::cerr << "Backup::slotCopyArchive: not isDir" << std::endl;
    }

    //std::cerr << "destUrl: " << destUrl.path() << std::endl;

    KIO::Job* job = KIO::file_move( mTmpArchiveUrl, destUrl, -1, true );
    connect(
        job, SIGNAL( result( KIO::Job * ) ),
             SLOT( slotFinished( KIO::Job *) )
        );
}

void Backup::slotFinished( KIO::Job* aJob )
{
    //std::cerr << "Backup::slotFinished" << std::endl;
    if ( aJob->error() )
    {
        aJob->showErrorDialog();
        return;
    }
    //std::cerr << "mTmpArchiveUrl: " << mTmpArchiveUrl.url() << std::endl;
    Q_ASSERT( !QFileInfo( mTmpArchiveUrl.path() ).exists() );
    Q_ASSERT( QFileInfo( mArchiveUrl.path() ).exists() );
}

KURL  Backup::defaultArchiveName( KURL source, KURL archive )
{
    // Create default file name:
    // $HOME/[filename|dirname]-yyyyMMddhhmmss.tar.gz
    QDateTime qdt = QDateTime::currentDateTime();
    QString filename = "file:";

    if ( archive.isEmpty() )
        filename += Helper::homeDir();
    else
        // HACK: Hier wird wohl noch der / am Ende hngen
        filename += archive.path();

    if ( filename.isNull() )
    {
        if ( KMessageBox::Cancel ==
             KMessageBox::warningContinueCancel(
                 0L, i18n(
                     "No $HOME was found. Backup will be saved in /tmp" ))
            )
        {
            return KURL();
        }
        else
        {
            filename += "/tmp";
        }
    }
    filename += "/" + source.fileName() + "-" +
                qdt.toString( "yyyyMMddhhmmss" ) + ".tar.gz";
    return KURL( filename );
}

KURL Backup::dotCheck( KURL inUrl )
{
    QString url_string = inUrl.url();
    int index = url_string.findRev( "/" );

    if ( -1 == index )
    {
        if ( url_string.startsWith( "." ) )
            return KURL( "dot" + url_string );
        else
            return inUrl;
    }
    if ( -1 < index )
    {
        QString rightstring = url_string.right( url_string.length() - index - 1 );
        QString leftstring  = url_string.left( index + 1 );

        if ( rightstring.startsWith( "." ) )
            return KURL( leftstring + "dot" + rightstring );
    }

    return inUrl;
}

////////////////////////////////////////////////////////////////////////////////
///////                  class BackupProfile                             ///////
////////////////////////////////////////////////////////////////////////////////

BackupProfile::BackupProfile( Backup*      backup,
                              QString      identifier,
                              TimeInterval interv,
                              QDateTime    last,
                              bool         activ,
                              bool         readonly )
    : Dumpable( "BackupProfile", identifier ),
      mIdentifier( identifier ),
      mTimeToBackup( interv ),
      mLastBackupTime( last ),
      mbIsActivated( activ ),
      mReadOnly( readonly )
{
    mpBackup = backup;
    mpTimer = new QTimer();
}

BackupProfile::~BackupProfile()
{
    Q_ASSERT( mpBackup );
    Q_ASSERT( mpTimer );

    delete mpBackup;
    delete mpTimer;
}

void BackupProfile::activate()
{
    if ( mReadOnly )
        return;
    if ( mbIsActivated )
        setupTimer();
}

void BackupProfile::restore()
{
    QFileInfo q( mpBackup->getArchiveUrl().path() );
    KURL      toRestore;

    if ( ! restoreActionIsSave() )
        return;

    if ( q.isDir() )
        toRestore = getToRestoreBackup();
    else
        toRestore = mpBackup->getArchiveUrl();

    KArchive *archive = new KTar( toRestore.path() );
    if ( ! archive->open( IO_ReadOnly ) )
    {
        KMessageBox::error( 0L, i18n( "Sorry, somethings wrong with your archive."
                                      "I cannot restore the backup" ) );
        return;
    }
    KArchiveEntry *root = const_cast<KArchiveDirectory*>( archive->directory() );
    this->extractArchiveEntry( root, QString::null );
    archive->close();
}

QString BackupProfile::dumpInternal()
{
    QString tmp;
    tmp += QString( "mpTimer" ) + QString().setNum( (long) this->mpTimer ) + "\n";
    tmp += QString( "Identifier:  " ) + this->mIdentifier + "\n";
    tmp += this->mTimeToBackup.dump();
    tmp += QString( "LastBackup:  " ) + this->mLastBackupTime.toString(
        "yyyyMMddhhmmss" ) + "\n";

    if ( mbIsActivated ) tmp += QString( "mbIsActivated:     true\n" );
    else                 tmp += QString( "mbIsActivated:     false\n" );
    if ( mReadOnly )     tmp += QString( "Readonly:          true\n" );
    else                 tmp += QString( "Readonly:          false\n" );

    tmp += mpBackup->dump();
    return tmp;
}

BackupProfile * BackupProfile::newBackupProfile(
    KURL      sourceUrl,
    KURL      backupUrl,
    QString   identifier,
    int       time,
    QString   measurement,
    bool      activ,
    QDateTime last
    )
{
    // XXX Check wether last is valid
    TimeInterval  t( (unsigned int) time, measurement );
    Backup *backup = new Backup(
        sourceUrl, backupUrl
        );
    return new BackupProfile( backup, identifier, t, last, activ );
}

BackupProfile *BackupProfile::newBackupProfileReadOnly(
    KURL      sourceUrl,
    KURL      backupUrl,
    QString   identifier,
    int       time,
    QString   measurement,
    bool      activ,
    QDateTime last
    )
{
    TimeInterval  t( (unsigned int) time, measurement );

    Backup *backup = new Backup(
        sourceUrl, backupUrl, true
        );
    return new BackupProfile(
        backup, identifier, t, last, activ, true
        );
}

BackupProfile *BackupProfile::newBackupProfileReadOnly( const BackupProfile * const b )
{
    TimeInterval  t( b->getTimeBetweenBackups() );
    Backup  *backup     = new Backup(
        b->getSourceUrl(), b->getArchiveUrl(), true
        );
    return new BackupProfile(
        backup, b->getIdentifier(), t, b->getLastTime(), b->isActivated(), true
        );
}

void BackupProfile::slotDoBackup()
{
    slotDoOneTimeBackup();

    if ( mbIsActivated )  // Important: this one after mLastBackupTime = ...
        setupTimer();
}

void BackupProfile::slotDoOneTimeBackup()
{
    if ( mReadOnly )
        return;

    mpBackup->doBackup();
    mLastBackupTime = QDateTime::currentDateTime();
}

BackupProfile::BackupProfile( const BackupProfile& b )
    : QObject(),
      Dumpable( "BackupProfile", b.mIdentifier + " copy" ),
      mIdentifier( b.mIdentifier ),
      mTimeToBackup( b.mTimeToBackup ),
      mLastBackupTime( b.mLastBackupTime ),
      mbIsActivated( false ),
      mReadOnly( true )
{
    mpTimer = new QTimer();
}

void BackupProfile::operator=( const BackupProfile& )
{
    Q_ASSERT( false );
}

void BackupProfile::setupTimer()
{
    int msecs = 0;

    if ( mpTimer->isActive() )
        mpTimer->stop();

    if ( !mLastBackupTime.isValid() )
        msecs = 10; // Do first backup nearly immediately // XXX: replace
    else if ( mLastBackupTime.addSecs( mTimeToBackup.secs() )
              <=
              QDateTime::currentDateTime() )
        msecs = 10; // Do first backup nearly immediately // XXX: replace

    else
        // Number of milliseconds from now to the next backup time
        msecs = 1000 * QDateTime::currentDateTime().secsTo(
            mLastBackupTime.addSecs( mTimeToBackup.secs() )
            ) ;
    mpTimer->singleShot( msecs, this, SLOT( slotDoBackup()));
}

void BackupProfile::extractArchiveEntry( KArchiveEntry* entry, QString path )
{
    if ( entry->isDirectory() )
    {
        QDir dir;
        if ( dir.exists( path ) || QString::null == path )
            /* NOOP */;
        else
        {
            if ( ! dir.mkdir( path ) )
                KMessageBox::error(
                    0L, i18n( "Shit happend: Could not create dir: %1" ).arg( path )
                    );
        }

        QStringList list = ( dynamic_cast<KArchiveDirectory*>( entry ) )->entries();
        QStringList::iterator it = list.begin();
        while ( list.end() != it )
        {
            extractArchiveEntry(
                ( dynamic_cast<KArchiveDirectory*>( entry ) )->entry( *it ),
                path + "/" + *it
                );
            it++;
        }
    }
    else if ( entry->isFile() )
    {
        QFile file( path );
        if ( !file.open( IO_WriteOnly ) )
            // XXX Once again a location for KNotify (is it?)
            KMessageBox::error(
                0L, i18n( "Shit happend! Could not open file: %1" ).arg( path )
                );
        else
        {
            file.writeBlock( ( dynamic_cast<KArchiveFile*>( entry ) )->data() );
            file.close();
        }
    }
    else
        // XXX see XXX above
        KMessageBox::error(
            0L, i18n( "Shit happend! Entry not dir nor file!" )
            );
}

KURL BackupProfile::getToRestoreBackup()
{
    QFileInfo qq( mpBackup->getArchiveUrl().path() );
    Q_ASSERT( qq.isDir() );

    QDir d = qq.dir();
    // XXX tar.gz or tar.bz2 or ... is missing

    QStringList list = d.entryList(
        mpBackup->getSourceUrl().fileName() + "-??????????????.tar.gz",
        QDir::Files,
        QDir::Name | QDir::Reversed
        );
    return *new KURL( mpBackup->getArchiveUrl().url() + *list.begin() );
}

bool BackupProfile::restoreActionIsSave()
{
    QFileInfo q( mpBackup->getSourceUrl().path() );

    if ( q.exists() )
    {
        KMessageBox::sorry(
            0L, i18n( "The url " + mpBackup->getSourceUrl().path() + " to be restored "
                      "still exists.\nPlease remove or rename ist first."),
            i18n( "Url exists" )
            );
        return false;
    }
    return true;
}
